/************************************************************************
 *                                                                      *
 *  Program package         U F F                                       *
 *                                                                      *
 *  Version   1.2                                                       *
 *  Date:     Fri Nov  1 11:11:59 1996                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'tr@uninova.pt'.                                     *
 *                                                                      *
 ************************************************************************/

/*
 *  Module to calculate Fourier descriptors
 *  refer e.g. to
 *  Gonzalez / Wintz : Digital Image Processing II. edition, page 404,
 *  Addison Wesley
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "def.h"
#include "unldef.h"

extern int N;
extern int NPoints;
extern CPoint *Points;
extern bool verbose;
extern bool trace;
extern bool outAscii;
extern str100 featNameFile;
extern str20 pr;


float *FD = NULL;
int FD_buf_len = 1024;
int FD_fraction = 8;

#ifdef PI
#undef PI
#endif
#define PI 3.14159265358979323846
#ifdef PI2
#undef PI2
#endif
#define  PI2  6.2831853071795862320
#define SQRT2 1.4142135623730950488


#define NOINVERSE 0
#define INVERSE 1

static char buf[20];
static float maxDist;
static CartesianCoord O;


#define SCALE (1000.0)	/* scale Fourier descriptors to a reasonable size */

static int nrFDs;	/* F0 and F1 are included but cannot be used */
static CPoint *pix = NULL;
static double *arclen = NULL;	/* accumulative arc length of the curve */
static CartesianCoord *pNormed = NULL;
static double *FFTBufReal = NULL, *FFTBufImag = NULL;
static bool FD_initialized = FALSE;
static int maxPoints;


/* #define CALC_EXPLICIT	/**/
#ifdef CALC_EXPLICIT
/* -----------------------------------------------------------------------
 *
 * For the calculus of the Fourier descriptors by definition
 *
*/
static float *FD_by_definition = NULL;
/*
 *  The lookup tables for the explicit Fourier transform
*/
static double *CosTab = NULL, *SinTab = NULL;
static void init_FD_explicit()
{
 int i;
 double angle = 0.0, omega = PI2 / (double)FD_buf_len;

 for( i = 0; i < FD_buf_len; i++ )
 {
   CosTab[i] = cos( angle );
   SinTab[i] = sin( angle );
   /* printf("angle=%lf   CosTab[%d]=%lf   SinTab[%d]=%lf\n",
		angle,i,CosTab[i],i,SinTab[i]);	/**/
   angle += omega;
 }
}


static float calc_one_FD( xCoord, yCoord, u )
double *xCoord, *yCoord;
int u;
{
 double sine, cosine, reX, imX, Re = 0.0, Im = 0.0;
 float fd;
 int i, indx;

 for( i = 0; i < FD_buf_len; i++ )
 {
   indx = ( u * i ) % FD_buf_len;
   cosine = CosTab[indx]; sine = SinTab[indx];
   reX = xCoord[i]; imX = yCoord[i];
   Re += reX * cosine + imX * sine;
   Im -= reX * sine - imX * cosine;
 }
 Re /= (double)FD_buf_len; Im /= (double)FD_buf_len;
 fd = (float)sqrt(Re*Re+Im*Im);
 return( fd );
}
#endif
/* ----------------------------------------------------------------------- */


void Alloc_FD()
{
 size_t dim;

 nrFDs = FD_buf_len;
 maxPoints = N*N;
 dim = (size_t) FD_buf_len;

 pNormed = (CartesianCoord*) malloc( dim * sizeof(CartesianCoord) );
 CHKPTR(pNormed);
 FFTBufReal = (double*) malloc( dim * sizeof(double) ); CHKPTR(FFTBufReal);
 FFTBufImag = (double*) malloc( dim * sizeof(double) ); CHKPTR(FFTBufImag);
 pix = (CPoint*) malloc( maxPoints * sizeof(CPoint) ); CHKPTR(pix);
 arclen = (double*) malloc( maxPoints * sizeof(double) ); CHKPTR(arclen);
 FD = (float*) malloc( dim * sizeof(float) ); CHKPTR(FD);

#ifdef CALC_EXPLICIT
 FD_by_definition = (float*) malloc( dim * sizeof(float) );
 CHKPTR(FD_by_definition);
 CosTab = (double*) malloc( dim * sizeof(double) ); CHKPTR(CosTab);
 SinTab = (double*) malloc( dim * sizeof(double) ); CHKPTR(SinTab);
#endif
}


void Free_FD()
{
 FREE(pNormed);
 FREE(FFTBufReal); FREE(FFTBufImag);
 FREE(pix); FREE(arclen);
 FREE(FD);

#ifdef CALC_EXPLICIT
 FREE(FD_by_definition);
 FREE(CosTab); FREE(SinTab);
#endif
}


/*
 *  Neighborhood mask for curve tracking
 -----------
| 6 | 5 | 4 |
|---|---|---|
| 7 | x | 3 |
|---|---|---|
| 0 | 1 | 2 |
 -----------
*
*/

#define NRNEI 8
/* counterclockwise direction */
static CPoint go[NRNEI] = {{1,0},{1,0},{0,1},{0,1},{-1,0},{-1,0},{0,-1},{0,-1}};

static int determine_from( fromX, fromY, srchX, srchY )
int fromX, fromY, srchX, srchY;
{
 if( fromX+1 == srchX && fromY+1 == srchY ) return( 0 );
 if( fromX == srchX && fromY+1 == srchY ) return( 1 );
 if( fromX-1 == srchX && fromY+1 == srchY ) return( 2 );
 if( fromX-1 == srchX && fromY == srchY ) return( 3 );
 if( fromX-1 == srchX && fromY-1 == srchY ) return( 4 );
 if( fromX == srchX && fromY-1 == srchY ) return( 5 );
 if( fromX+1 == srchX && fromY-1 == srchY ) return( 6 );
 if( fromX+1 == srchX && fromY == srchY ) return( 7 );
 fprintf( stderr, "Error... exit\n"); exit( 1 );
}


/* The Arclength between two pixels
   4 Neighbors = 1
   diagonal Neighbors = Sqrt(2)
*/
static float arc_len_neib[NRNEI] = { SQRT2,1.0,SQRT2,1.0,SQRT2,1.0,SQRT2,1.0 };


static void EuclidDist( P1, P2, D )
CPoint *P1;
CartesianCoord *P2;
float *D;
{
 float x, y;

 x = P1->x - P2->x;
 y = P1->y - P2->y;
 *D = sqrt( x*x + y*y );
}

static void calcCentroid( P )
Pixel **P;
{
 int i, x, y;
 float dist;

 O.x = 0.0; O.y = 0.0; NPoints = 0;
 for( x = 0; x < N; x++ )
   for( y = 0; y < N; y++ )
   {
     if ( P[x][y] )
     {
       O.x = O.x + x; O.y = O.y + y;
       Points[NPoints].x = x;  Points[NPoints].y = y;
       NPoints += 1;
     }
   }
 /* printf("Number of Points in original Pattern: %d", NPoints); DBG; /**/
 O.x /= (float)NPoints;
 O.y /= (float)NPoints;
 maxDist = 0.0;
 for ( i = 0; i < NPoints; i++ )
 {
   EuclidDist(&(Points[i]), &O, &dist);
   if ( dist > maxDist)
     maxDist = dist;
 }
}




/*
 *	fft1D.c
 *
 *	C Version 1.0 by Steve Sampson, Public Domain
 *
 *	This program is based on the work by W. D. Stanley
 *	and S. J. Peterson, Old Dominion University.
 *
 */

#define	TWO_PI	6.2831853071795862320

static int power;

/*
 *	Bit reverse the number
 *
 *	Change 11100000b to 00000111b or vice-versa
 */
static int permute(index, len)
int	index, len;
{
	int	n1, result, loop;

	n1 = len;
	result = 0;
	for (loop = 0; loop < power; loop++)  {
		n1 >>= 1;			/* n1 / 2.0 */
		if (index < n1)
			continue;
		result += (int) pow((double)2.0, (double)loop);
		index -= n1;
	}
	return result;
}


static void fft1D( RE, IM, len )
double *RE, *IM;
int len;
{
	unsigned i1, i2, i3, i4, y;
	int	 i, loop, loop1, loop2;
	double	 a1, a2, b1, b2, z1, z2, v, *tmpBuf = NULL;

	power =  (int) (log10((double)len) / log10(2.0));
	/* Scale the data */
	for (loop = 0; loop < len; loop++)  {
		RE[loop] /= (double)len;
		IM[loop] /= (double)len;
	}
	i1 = len >> 1;
	i2 = 1;
	v = TWO_PI * ((double)1.0 / (double)len);

	for (loop = 0; loop < power; loop++)  {
		i3 = 0;
		i4 = i1;
		for (loop1 = 0; loop1 < i2; loop1++)  {
			y = permute(i3 / i1, len);
			z1 =  cos(v * y);
			z2 = -sin(v * y);
			for (loop2 = i3; loop2 < i4; loop2++)  {
				a1 = RE[loop2];
				a2 = IM[loop2];

				b1 = z1*RE[loop2+i1] - z2*IM[loop2+i1];
				b2 = z2*RE[loop2+i1] + z1*IM[loop2+i1];

				RE[loop2]      = a1 + b1;
				IM[loop2]      = a2 + b2;

				RE[loop2 + i1] = a1 - b1;
				IM[loop2 + i1] = a2 - b2;
			}
			i3 += (i1 << 1);
			i4 += (i1 << 1);
		}
		i1 >>= 1;
		i2 <<= 1;
	}
	/* bit reverse the indices */
	tmpBuf = (double*) malloc( len * sizeof(double) );
	if( tmpBuf == NULL )
            { fprintf(stderr, "FFT1D: cannot allocate 'tmpBuf' - Exit...\n");
		exit(1); }
	for( i = 0; i < len; i++ )
		tmpBuf[i] = RE[i];
	for( i = 0; i < len; i++ )
		RE[permute(i,len)] = tmpBuf[i];
	for( i = 0; i < len; i++ )
		tmpBuf[i] = IM[i];
	for( i = 0; i < len; i++ )
		IM[permute(i,len)] = tmpBuf[i];
	free(tmpBuf); tmpBuf = NULL;
}

#define EPS 0.00001

static void NormPoints( nrpix, total_arclen )
int nrpix;
double total_arclen;
{
 double res, dx, dy, dl, t, x, y, xi, yi;
 int i, L, Ll, Lr;

 if( FD_buf_len == 1)
 {
   fprintf( stderr, "Resolution 1 does not make sense. Error. Exit...\n");
   exit(1);
 }

 /* Reset the buffer in order to check out if it has been filled */
 for( i = 0; i < FD_buf_len; i++ )
 {
   pNormed[i].x = (float)EMPTY; pNormed[i].y = (float)EMPTY;
 }

 /* Create a 'discrete' arc length */
 res = (double)(FD_buf_len-1) / total_arclen; 

 /* printf("Resolution of buffer=%d  nrpix=%d  total_arclen=%lf\n",
	FD_buf_len, nrpix, total_arclen );	/**/

 /* Normalize the arc length of the curve */
 /* printf("Normalizing arc length with factor %lf ...\n", res);	/**/
 for( i = 0; i < nrpix; i++ )
 {
   /* printf("i=%3d Point =(%3d,%3d) Arc length: Before: %6.2lf  After: ",
	i, pix[i].x, pix[i].y, arclen[i]);	/**/
   arclen[i] *= (float)res;
   /* printf("%6.2lf\n", arclen[i]);   /**/
 }


 for( i = 0; i < nrpix - 1; i++ )
 {
   dx = (double)(pix[i+1].x - pix[i].x); dy = (double)(pix[i+1].y - pix[i].y);
   xi = (double)pix[i].x; yi = (double)pix[i].y;
   dl = arclen[i+1] - arclen[i];

   Ll = (int) floor ( arclen[i] );
   Lr = (int) ceil ( arclen[i+1] );

   for( L = Ll; L <= Lr; L++ )
   {
     t = ( (double)L - arclen[i] ) / dl;
     if( t >= 0.0-EPS && t <= 1.0+EPS )
     {
       x = xi + t * dx; y = yi + t * dy;
       pNormed[L].x = (float)x, pNormed[L].y = (float)y;

       /* printf("i=%3d (%3d,%3d)---(%3d,%3d) t=%.2lf buf[%3d]=(%.3lf,%.3lf)\n",
		i, pix[i].x, pix[i].y, pix[i+1].x, pix[i+1].y, t, L, x, y); /**/
     }
     else
     {
       /* printf("i=%3d --> (%3d,%3d)---(%3d,%3d) t=%.2lf   *** Failing...\n",
		i, pix[i].x, pix[i].y, pix[i+1].x, pix[i+1].y, t );	/**/
     }
   }
 }
 pNormed[FD_buf_len-1].x = (float)(pix[nrpix-1].x);
 pNormed[FD_buf_len-1].y = (float)(pix[nrpix-1].y);

 /* Check if everything has been filled in properly */
 for( i = 0; i < FD_buf_len; i++ )
 {
   /* printf("Fourier buffer[%3d]=(%.3f,%.3f)\n",
		i, pNormed[i].x, pNormed[i].y );	/**/
   if( pNormed[i].x == (float)EMPTY ||
       pNormed[i].y == (float)EMPTY )
   {
     fprintf( stderr, "Could not fill the array. Error. Exit...\n");
     exit(1);
   }
 }
}


#define FEW_PIX 10

void calc_1D_Fourier_Descriptors( P )
Pixel **P;
{
 int x, y, firstX, firstY, fromX, fromY, srchX, srchY, nrpix;
 int xMin, xMax, yMin, yMax;
 bool found, around;
 int from, next, srchCnt, i;
 double RE, IM, F0, F1, mag;
 double acc_arclen;

#ifdef CALC_EXPLICIT
 /* init at first time */
 if( ! FD_initialized )
 {
   init_FD_explicit();
   FD_initialized = TRUE;
 }
#endif

 calcCentroid( P );

 /* Set the bounding box of the pattern */
 xMin = ROUND(O.x - maxDist) - 2;
 if( xMin < 0 )
   xMin = 0;
 xMax = ROUND((O.x + maxDist)) + 2;
 if(xMax > (N-1))
   xMax= N-1;
 yMin = ROUND((O.y - maxDist)) - 2;
 if(yMin < 0)
   yMin= 0;
 yMax = ROUND((O.y + maxDist)) + 2;
 if(yMax > (N-1))
   yMax= N-1;

 /* search the first point */
 found = FALSE; x = xMin; y = yMin;
 while( ! found )
 {
   while( ! found && x < xMax )
   {
     found = (int)P[x][y];
     if( ! found )
       x++;
   }
   if( ! found )
     { x = xMin; y++; }
 }
 if( ! found ) EXIT("Could not find first pixel of pattern");

 if( verbose )
   printf("\n%sFound first pixel at (%d,%d)\n", pr, x, y );
 nrpix = 0;
 firstX = x; firstY = y;
 fromX = x-1; fromY = y;
 from = 7;

 acc_arclen = 0.0;
 pix[ nrpix ].x = firstX; pix[ nrpix ].y = firstY;
 arclen[ nrpix ] = acc_arclen;
 nrpix++;

 around = FALSE;
 while( ! around )
 {
   /* search next pixel in counterclockwise direction */
   found = FALSE; next = from; srchCnt = 1;
   srchX = fromX + go[from].x; srchY = fromY + go[from].y;
   while( ! found )
   {
     found = (int)P[srchX][srchY];
     /* printf("searching at (%d,%d)  found=%d ",
	 srchX, srchY, found ); /**/
     if( ! found )
     {
       srchCnt++;
       if( srchCnt > NRNEI )
       {
         /* turned around too much = pixel is isolated */
         printf("Cannot find next pixel. Exit..."); DBG; exit(1);
       }
       next = (next+1) % NRNEI; 
       /* printf("next=%d   new_next=%d\n", next, (next+1) % NRNEI ); /**/
       srchX = srchX + go[next].x; srchY = srchY + go[next].y;
     }
   } 
   /* found the next pixel */
   /* determine the new "from" coordinate */
   from = determine_from( x, y, srchX, srchY );

   /* printf("*** next pixel = (%3d,%3d) from direction = %d\n",
		srchX, srchY, from ); /**/

   pix[ nrpix ].x = srchX; pix[ nrpix ].y = srchY;

   /* Accumulate the arc length */
   acc_arclen += arc_len_neib[from];
   arclen[ nrpix ] = acc_arclen;
   /* printf("pix[%3d]=(%3d,%3d) Arclength=%.2lf Sum_arc_length[%d]=%.2lf\n",
	nrpix, pix[ nrpix ].x, pix[ nrpix ].y, arc_len_neib[from],
	nrpix, acc_arclen ); DBG;	/**/
   nrpix++;

   fromX = x; fromY = y;
   x = srchX; y = srchY;
   around = ( (srchX==firstX) && (srchY==firstY) );
 }
 /* The pixel n is again pixel 0 */
 if( verbose )
   printf("%sNumber of pixels (counting first pixel twice) = %d\n", pr, nrpix );
 if( nrpix >= maxPoints )
   { printf("Too many points. Exit...\n"); exit(1); }
 /* Issue a warning if only a few pixels were found */
 if( nrpix < FEW_PIX )
 {
   fprintf(stderr,"\nWARNING: Found less than %d pixels on the contour.\n",
	FEW_PIX);
   fprintf(stderr,"\tThis could mean the trace algorithm had been confused\n");
   fprintf(stderr,"\tby a one-pixel thick 'hair' emerging from the contour.\n");
   fprintf(stderr,"\tThis should be avoided by raising the resolution");
   fprintf(stderr,"of the image.\n");
   fprintf(stderr,"Exitus..."); gets(buf); exit(1);
 }

 /*--------------------------------------------------------------------*/
 /* Map the pixel sequence to the buffer which has a power-of-2 length */
 /* - Normalize with respect to the Arc Length of the curve            */
 /*--------------------------------------------------------------------*/
 NormPoints( nrpix, acc_arclen );

 /* Prepare the buffer with the Fourier coefficients */
 for( i = 0; i < FD_buf_len; i++ )
 {
   /* if( i < 10 || FD_buf_len - i < 10 )
     printf("pNormed[%d]=(%f,%f)\n", i, pNormed[i].x, pNormed[i].y ); /**/
   FFTBufReal[i] = (double)pNormed[i].x;
   FFTBufImag[i] = (double)pNormed[i].y;
 }

 
#ifdef CALC_EXPLICIT
 /*  Here the Fourier descriptors are calculated by definition */
 for( i = 0; i < FD_buf_len; i++ )
   FD_by_definition[i] = calc_one_FD( FFTBufReal, FFTBufImag, i );
 F1 = FD_by_definition[1];
 for( i = 0; i < FD_buf_len; i++ )
   FD_by_definition[i] = SCALE * FD_by_definition[i] / F1;
#endif


 fft1D( FFTBufReal, FFTBufImag, FD_buf_len );

 RE = FFTBufReal[ 1 ]; IM = FFTBufImag[ 1 ];
 F1 = sqrt( RE*RE + IM*IM );
 /* ignore FD[0] and the normalized FD[1] */
 for( i = 0; i < FD_buf_len; i++ )
 {
   RE = FFTBufReal[i]; IM = FFTBufImag[i];
   mag = sqrt( RE*RE + IM*IM );
   if( i > 1 && mag > F1 )
     { printf("Warning FD[%d] is larger than F1  (%lf>%lf)", i, mag, F1); DBG; }
   FD[ i ] = SCALE * mag / F1;
 }
}
